import java.awt.Point;
import java.util.Vector;

/**
 * AStar. Classe qui gre le "pathfinding"
 *
 * @author      Andreas Jaggi, Sebastian Gfeller
 * @created     7. Mai 2004
 * @version     AStar.java,v 1.7 2004/06/23 11:34:33 ajaggi Exp
 * @copyright   Andreas Jaggi, Sebastian Gfeller 2004
 * @license     http://opensource.org/licenses/gpl-license.php GNU Public License
 */

public class AStar implements Constantes
{

	/**
	 * Vecteur contenant tous les Blocks  contrler
	 */
	protected Vector ouvert;
	
	/**
	 * Vecteur contenant tous les Blocks contrls
	 */
	protected Vector controle;

	/**
	 * Vecteur dans lequel on sauvegarde le chemin trouv
	 */
	private Vector chemin = new Vector();

	/**
	 * Coordonnes de l'arrive
	 */
	private Point targetCoord;

	/**
	 * Coordonnes du dpart
	 */
	private Point startCoord;

	/**
	 * Le plateau du jeu
	 */
	private Plateau plateau;

	/**
	 * Le niveau du jeu
	 */
	private Niveau niveau;

	/**
	 * Nombre maximale des tests qu'on fait pour chercher un chemin 
	 * (on n'a pas besoin de contrler plus de Blocks qu'il y a sur la carte)
	 */
	private int maxPas = NBLIGNES*NBCOLONNES;

	/**
	 * Cout minimal d'un Block, utilis pour stimer les couts d'un Block
	 */
	private double minCout;

	/**
	 * Constructeur de AStar
	 *
	 * @param plateau Plateau du jeu
	 * @param niveau Niveau du jeu
	 */
	public AStar ( Plateau plateau, Niveau niveau ) {
		/* cherger les rfrences */
		this.plateau = plateau;
		this.niveau = niveau;
	}

	/**
	 * Calcule le chemin entre deux points
	 *
	 * @param start Point de dpart
	 * @param target Point d'arrive
	 *
	 * @return Le chemin
	 */
	public boolean calculeChemin ( Point start, Point target ) {

		if ( DEBUG ) {
			System.out.println("Setting start ("+start.x+", "+start.y+")");
		}

		/* enregistre le dpart */
		this.startCoord = start;

		if ( DEBUG ) {
			System.out.println("Setting target ("+target.x+", "+target.y+")");
		}

		/* enregistre le fin */
		this.targetCoord = target;

		if ( DEBUG ) {
			System.out.println("Updating map");
		}

		/* actualise la carte */
		this.niveau.resetBlocks();

		if ( DEBUG ) {
			System.out.println("Searching path");
		}

		/* cherche un chemin */
		return this.trouveChemin();
	}

	/**
	 * Calcule une estimation des couts pour aller d'un point  un autre
	 *
	 * @param a Point de dpart
	 * @param b Point d'arrive
	 * @param coeff Cofficient (e.g. le cout estim d'un Block)
	 *
	 * @return Distance
	 */
	public double cbDist ( Point a, Point b, double coeff ) {

		/* calcule de la distance relle */
		double dx1 = a.x - b.x;
		double dy1 = a.y - b.y;

		/* car la distance relle donne beaucoup des distances du mme valeur,
		 * on ajoute un facteur "distingant" */
		double dx2 = this.startCoord.x - b.x;
		double dy2 = this.startCoord.y - b.y;
		double det = Math.abs( dx1 * dy2 - dx2 * dy1 );

		return coeff * ( Math.abs(dx1) + Math.abs(dy1) + det * 0.0002 );
	}

	/**
	 * Trouve un chemin
	 *
	 * @return startCoord et targetCoord sont valables
	 */
	public boolean trouveChemin ( ) {

		/* cherche le cout le plus bas dans la carte et l'assigne  minCout */
		this.minCout = Double.MAX_VALUE;

		for ( int i=0; i < NBLIGNES; i++ ) {
			for ( int j=0; j < NBCOLONNES; j++ ) {
				this.minCout = Math.min(this.niveau.carte[i][j].getCouts(), this.minCout);
			}
		}

		/* intialisation des Vecteurs fondamentaux du AStar */
		this.ouvert = new Vector();
		this.controle = new Vector();

		/* si le dpart n'existe pas, on casse */
		if ( this.niveau.carte[this.startCoord.x][this.startCoord.y] == null ) {
			return false;
		}

		/* si l'arrive n'existe pas, on casse */
		if ( this.niveau.carte[this.targetCoord.x][this.targetCoord.y] == null ) {
			return false;
		}

		if ( DEBUG ) {
			System.out.println("Starting search engine");
		}

		/* commence la recherche d'un chemin */
		return this.start();
	}

	/**
	 * La fonction boucle du AStar
	 */
	public boolean start ( ) {
		/* on commence par ajouter le Block de dpart aux Vecteur  contrler */
		this.ouvert.addElement(this.niveau.carte[this.startCoord.x][this.startCoord.y]);

		/* initilisation des variables locaux */
		int pas = 0;
		int status = 0;

		/* apple this.pas() jusqu'il trouve un chemin ou on atteind le nombre maximal des pas admis */
		while ( status == 0 && pas < this.maxPas ) {
			pas++;
			status = this.pas();
		}

		/* on a trouv un chemin */
		if ( status == 1 ) {

			if ( DEBUG ) {
				System.out.println("Setting Path");
			}

			/* on sauvegarde le chemin trouv */
			this.setChemin();
			return true;

		/* pas de chemin trouvee */
		} else {
			if ( DEBUG ) {
				System.out.println("No Path found");
			}

			return false;
		}
	}

	/**
	 * Contrle un Block
	 *
	 * @return Status: 1 = chemin trouve, 0 = pas encore trouve, -1 = pas de chemin possible
	 */
	public int pas ( ) {
		/* on charge l'arrive */
		Block fin = this.niveau.carte[this.targetCoord.x][this.targetCoord.y];

		/* initialisation des variables locaux */
		double min = Double.MAX_VALUE;
		double score;
		Block best = (Block)this.ouvert.elementAt(this.ouvert.size()-1);
		Block actuel;

		/* choisis le block qui est le plus proche au target */

		for ( int i=0; i < this.ouvert.size(); i++ ) {
			actuel = (Block)this.ouvert.elementAt(i);

			/* calcule la distance seulement si on ne l'a pas encore fait */
			if ( !this.controle.contains(actuel) ) {
				/* distance jusqu'ici */ 
				score = actuel.getDistDuStart(this.niveau.carte[this.startCoord.x][this.startCoord.y]);
				/* distance stime jusqu'au fin */
				score += this.cbDist(actuel.getPosition(), this.targetCoord, this.minCout);
				/* si le Block actuel possde un meilleur chemin au fin, on le met dans best */
				if ( score < min ) {
					min = score;
					best = actuel;
				}
			}
		}

		/* on prend le Block qui possde le meilleur chemin au fin */
		actuel = best;

		if ( DEBUG ) {
			System.out.println(actuel.getPosition().x+", "+actuel.getPosition().y);
		}

		/* on enleve le Block du liste des Blocks  controler */
		this.ouvert.removeElement(actuel);

		/* et on le met dans la liste des Blocks controls */
		this.controle.addElement(actuel);

		/* on cherche les voisins du Block */
		Block voisins[] = this.niveau.getVoisins(actuel);

		/* contrle tous les voisins */

		for( int i = 0; i < 4; i++ ) {
			/* s'il y a un voisin  la position i ... */
			if ( voisins[i] != null ) {
				/* si on est arrive au fin, on quitte avec 1 */
				if ( voisins[i] == fin ) {
					if ( DEBUG ) {
						System.out.println("Found Path!");
					}

					return 1;
				}
				/* on contrle seulement des Block traversables */
				if ( !voisins[i].isIntraversable() ) {
					/* on dit au Block voisin i avec quel longeur du chemin on l'a trouv */
					voisins[i].ajouteADistDuStart(actuel.getDistDuStart(this.niveau.carte[this.startCoord.x][this.startCoord.y]));

					/* si on n'a pas encore contrle le Block et si on ne l'a pas encore mis sur la liste  contrler, on l'y ajoute */
					/* ... Je n'ai pas peur de la route
					 * Faudra voir, faut qu'on y goute ... [Noir Dsir - Le vent nous portera] */
					if ( !this.ouvert.contains(voisins[i]) && !this.controle.contains(voisins[i]) ) {
						this.ouvert.addElement(voisins[i]);
					}
				}
			}
		}

		/* on a controle tous les Blocks accessible (-> la liste  contrler n'a plus d'lements)
		 * alors on quitte avec -1 */
		if ( this.ouvert.size() == 0 ) {
			return -1;
		}

		/* sinon on quitte avec 0 */
		return 0;
	}

	/**
	 * Sauvegarde le chemin trouv dans this.chemin
	 */
	public void setChemin() {
		/* initialisation des variables */
		boolean fini = false;
		this.chemin = new Vector();
		Block prochain;
		/* on commence par l'arrive ..*/
		Block actuel = this.niveau.carte[this.targetCoord.x][this.targetCoord.y];
		/* .. et va jusqu'au dpart */
		Block stop = this.niveau.carte[this.startCoord.x][this.startCoord.y];
		int pas = 0;

		if ( DEBUG ) {
			System.out.println(this.chemin.size());
		}

		/* pendant qu'on n'est pas arrive au dpart et qu'on n'as pas contl tous les Blocks ... */
		while ( !fini && pas < this.maxPas ) {

			/* si le Block actuel n'est pas l'arrive,
			 * on l'ajoute au chemin */
			if ( actuel.getPosition() != this.targetCoord ) {
				this.chemin.addElement(actuel);
			}

			/* on cherche le Block le plus proche au dpart parmi les voisins du Block actuel */
			prochain = this.niveau.getPlusPetitVoisin(actuel, this.startCoord);
			/* ceci donne le nouveau Block actuel */
			actuel = prochain;

			if ( DEBUG ) {
				System.out.println(this.chemin.size());
			}

			/* si on est arrive au dpart, on met la condition de terminaison du boucle */
			if ( actuel == stop ) {
				fini = true;
				break;
			}

			pas++;
		}
	}

	/**
	 * Donne le chemin
	 * @return Le chemin
	 */
	public Vector getChemin ( ) {
		return (Vector)(this.chemin);
	}
}
